@@ -41,19 +41,19 @@
 /**
  * GdkEvent: (ref-func gdk_event_ref) (unref-func gdk_event_unref)
  *
- * `GdkEvent`s are immutable data structures, created by GDK to
- * represent windowing system events.
+ * Represents windowing system events.
  *
  * In GTK applications the events are handled automatically by toplevel
  * widgets and passed on to the event controllers of appropriate widgets,
  * so using `GdkEvent` and its related API is rarely needed.
+ *
+ * `GdkEvent` structs are immutable.
  */
 
 /**
  * GdkEventSequence:
  *
- * `GdkEventSequence` is an opaque type representing a sequence
- * of related touch events.
+ * An opaque type representing a sequence of related events.
  */
 
 static void
@@ -2890,6 +2890,7 @@ GDK_DEFINE_EVENT_TYPE (GdkPadEvent, gdk_pad_event,
                        GDK_EVENT_TYPE_SLOT (GDK_PAD_BUTTON_RELEASE)
                        GDK_EVENT_TYPE_SLOT (GDK_PAD_RING)
                        GDK_EVENT_TYPE_SLOT (GDK_PAD_STRIP)
+                       GDK_EVENT_TYPE_SLOT (GDK_PAD_DIAL)
                        GDK_EVENT_TYPE_SLOT (GDK_PAD_GROUP_MODE))
 
 GdkEvent *
@@ -2930,6 +2931,25 @@ gdk_pad_event_new_strip (GdkSurface *surface,
   return (GdkEvent *) self;
 }
 
+GdkEvent *
+gdk_pad_event_new_dial (GdkSurface *surface,
+                        GdkDevice  *device,
+                        guint32     time,
+                        guint       group,
+                        guint       index,
+                        guint       mode,
+                        double      value)
+{
+  GdkPadEvent *self = gdk_event_alloc (GDK_PAD_DIAL, surface, device, time);
+
+  self->group = group;
+  self->index = index;
+  self->mode = mode;
+  self->value = value;
+
+  return (GdkEvent *) self;
+}
+
 GdkEvent *
 gdk_pad_event_new_button (GdkEventType  type,
                           GdkSurface   *surface,
@@ -3005,7 +3025,8 @@ gdk_pad_event_get_axis_value (GdkEvent *event,
 
   g_return_if_fail (GDK_IS_EVENT (event));
   g_return_if_fail (GDK_IS_EVENT_TYPE (event, GDK_PAD_RING) ||
-                    GDK_IS_EVENT_TYPE (event, GDK_PAD_STRIP));
+                    GDK_IS_EVENT_TYPE (event, GDK_PAD_STRIP) ||
+                    GDK_IS_EVENT_TYPE (event, GDK_PAD_DIAL));
 
   *index = self->index;
   *value = self->value;
@@ -3031,7 +3052,8 @@ gdk_pad_event_get_group_mode (GdkEvent *event,
                     GDK_IS_EVENT_TYPE (event, GDK_PAD_BUTTON_PRESS) ||
                     GDK_IS_EVENT_TYPE (event, GDK_PAD_BUTTON_RELEASE) ||
                     GDK_IS_EVENT_TYPE (event, GDK_PAD_RING) ||
-                    GDK_IS_EVENT_TYPE (event, GDK_PAD_STRIP));
+                    GDK_IS_EVENT_TYPE (event, GDK_PAD_STRIP) ||
+                    GDK_IS_EVENT_TYPE (event, GDK_PAD_DIAL));
 
   *group = self->group;
   *mode = self->mode;
