@@ -21,7 +21,6 @@
 
 #include "gdkmemoryformatprivate.h"
 #include "gdkdisplay-wayland.h"
-#include "gdkprivate-wayland.h"
 #include "gdkdmabuftextureprivate.h"
 #include "gdksurface-wayland-private.h"
 #include "gdksubsurfaceprivate.h"
@@ -29,77 +28,45 @@
 #include "gdkglcontextprivate.h"
 #include "gdkgltextureprivate.h"
 #include "gsk/gskrectprivate.h"
+#include "gdkshm-private.h"
+
+#include "gdkdmabuffourccprivate.h"
 
 #include "linux-dmabuf-unstable-v1-client-protocol.h"
 
-G_DEFINE_TYPE (GdkWaylandSubsurface, gdk_wayland_subsurface, GDK_TYPE_SUBSURFACE)
+/* {{{ Utilities */
 
-static void
-gdk_wayland_subsurface_init (GdkWaylandSubsurface *self)
+/* Note: The GdkDihedral transforms are *inverses* of the corresponding
+ * wl_output_transform transforms.
+ *
+ * This is intentional: The GdkDihedral is the transform we want the
+ * compositor to apply. set_buffer_transform is about *already transformed*
+ * content. By telling the compositor that the content is already transformed
+ * by the inverse of the GdkDihedral, we get it to apply the transform we want.
+ */
+static inline enum wl_output_transform
+gdk_texture_transform_to_wl (GdkDihedral transform)
 {
+  return (enum wl_output_transform) transform;
 }
 
-static void
-gdk_wayland_subsurface_finalize (GObject *object)
+static inline GdkDihedral
+wl_output_transform_to_gdk (enum wl_output_transform transform)
 {
-  GdkWaylandSubsurface *self = GDK_WAYLAND_SUBSURFACE (object);
-
-  g_clear_object (&self->texture);
-  g_clear_pointer (&self->frame_callback, wl_callback_destroy);
-  g_clear_pointer (&self->opaque_region, wl_region_destroy);
-  g_clear_pointer (&self->viewport, wp_viewport_destroy);
-  g_clear_pointer (&self->color, gdk_wayland_color_surface_free);
-  g_clear_pointer (&self->subsurface, wl_subsurface_destroy);
-  g_clear_pointer (&self->surface, wl_surface_destroy);
-  g_clear_pointer (&self->bg_viewport, wp_viewport_destroy);
-  g_clear_pointer (&self->bg_subsurface, wl_subsurface_destroy);
-  g_clear_pointer (&self->bg_surface, wl_surface_destroy);
-  g_clear_pointer (&self->idle_inhibitor, zwp_idle_inhibitor_v1_destroy);
-
-  G_OBJECT_CLASS (gdk_wayland_subsurface_parent_class)->finalize (object);
+  return (GdkDihedral) transform;
 }
 
-static void
-shm_buffer_release (void             *data,
-                    struct wl_buffer *buffer)
-{
-  cairo_surface_t *surface = data;
+G_STATIC_ASSERT ((int) WL_OUTPUT_TRANSFORM_NORMAL == (int) GDK_DIHEDRAL_NORMAL);
+G_STATIC_ASSERT ((int) WL_OUTPUT_TRANSFORM_90 == (int) GDK_DIHEDRAL_90);
+G_STATIC_ASSERT ((int) WL_OUTPUT_TRANSFORM_180 == (int) GDK_DIHEDRAL_180);
+G_STATIC_ASSERT ((int) WL_OUTPUT_TRANSFORM_270 == (int) GDK_DIHEDRAL_270);
+G_STATIC_ASSERT ((int) WL_OUTPUT_TRANSFORM_FLIPPED == (int) GDK_DIHEDRAL_FLIPPED);
+G_STATIC_ASSERT ((int) WL_OUTPUT_TRANSFORM_FLIPPED_90 == (int) GDK_DIHEDRAL_FLIPPED_90);
+G_STATIC_ASSERT ((int) WL_OUTPUT_TRANSFORM_FLIPPED_180 == (int) GDK_DIHEDRAL_FLIPPED_180);
+G_STATIC_ASSERT ((int) WL_OUTPUT_TRANSFORM_FLIPPED_270 == (int) GDK_DIHEDRAL_FLIPPED_270);
 
-  /* Note: the wl_buffer is destroyed as cairo user data */
-  cairo_surface_destroy (surface);
-}
-
-static const struct wl_buffer_listener shm_buffer_listener = {
-  shm_buffer_release,
-};
-
-static struct wl_buffer *
-get_shm_wl_buffer (GdkWaylandSubsurface *self,
-                   GdkTexture           *texture)
-{
-  GdkWaylandDisplay *display = GDK_WAYLAND_DISPLAY (gdk_surface_get_display (GDK_SUBSURFACE (self)->parent));
-  int width, height;
-  cairo_surface_t *surface;
-  GdkTextureDownloader *downloader;
-  struct wl_buffer *buffer;
-
-  width = gdk_texture_get_width (texture);
-  height = gdk_texture_get_height (texture);
-  surface = gdk_wayland_display_create_shm_surface (display, width, height);
-
-  downloader = gdk_texture_downloader_new (texture);
-
-  gdk_texture_downloader_download_into (downloader,
-                                        cairo_image_surface_get_data (surface),
-                                        cairo_image_surface_get_stride (surface));
-
-  gdk_texture_downloader_free (downloader);
-
-  buffer = _gdk_wayland_shm_surface_get_wl_buffer (surface);
-  wl_buffer_add_listener (buffer, &shm_buffer_listener, surface);
-
-  return buffer;
-}
+/* }}} */
+/* {{{ Dmabuf buffer handling */
 
 static void
 dmabuf_buffer_release (void             *data,
@@ -220,6 +187,9 @@ get_dmabuf_texture_wl_buffer (GdkWaylandSubsurface *self,
                                g_object_ref (texture));
 }
 
+/* }}} */
+/* {{{ GL texture buffer handling */
+
 typedef struct {
   GdkTexture *texture;
   GdkDmabuf dmabuf;
@@ -245,26 +215,12 @@ static const struct wl_buffer_listener gl_buffer_listener = {
 
 static struct wl_buffer *
 get_gl_texture_wl_buffer (GdkWaylandSubsurface *self,
-                          GdkTexture           *texture)
+                          GdkTexture           *texture,
+                          GdkDmabuf            *dmabuf)
 {
-  GdkDisplay *display = gdk_surface_get_display (GDK_SUBSURFACE (self)->parent);
-  GdkGLTexture *gltexture = GDK_GL_TEXTURE (texture);
-  GdkGLContext *glcontext;
   GLBufferData gldata;
 
-  glcontext = gdk_display_get_gl_context (display);
-  if (glcontext == NULL ||
-      !gdk_gl_context_is_shared (glcontext, gdk_gl_texture_get_context (gltexture)))
-    return NULL;
-
-  /* Can we avoid this when a right context is current already? */
-  gdk_gl_context_make_current (glcontext);
-
-  if (!gdk_gl_context_export_dmabuf (glcontext,
-                                     gdk_gl_texture_get_id (gltexture),
-                                     &gldata.dmabuf))
-    return NULL;
-
+  gldata.dmabuf = *dmabuf;
   gldata.texture = g_object_ref (texture);
 
   return get_dmabuf_wl_buffer (self,
@@ -275,27 +231,102 @@ get_gl_texture_wl_buffer (GdkWaylandSubsurface *self,
                                g_memdup2 (&gldata, sizeof (gldata)));
 }
 
+static gboolean
+export_gl_texture_as_dmabuf (GdkDisplay *display,
+                             GdkTexture *texture,
+                             GdkDmabuf  *dmabuf)
+{
+  GdkGLTexture *gltexture = GDK_GL_TEXTURE (texture);
+  GdkGLContext *glcontext;
+
+  dmabuf->n_planes = 0;
+
+  glcontext = gdk_display_get_gl_context (display);
+  if (glcontext == NULL ||
+      !gdk_gl_context_is_shared (glcontext, gdk_gl_texture_get_context (gltexture)))
+    return FALSE;
+
+  /* Can we avoid this when a right context is current already? */
+  gdk_gl_context_make_current (glcontext);
+
+  return gdk_gl_context_export_dmabuf (glcontext,
+                                       gdk_gl_texture_get_id (gltexture),
+                                       dmabuf);
+}
+
+/* }}} */
+/* {{{ General texture buffer handling */
+
+static gboolean
+get_texture_info (GdkWaylandSubsurface *self,
+                  GdkTexture           *texture,
+                  guint32              *out_fourcc,
+                  gboolean             *out_premultiplied,
+                  GdkDmabuf            *out_dmabuf)
+{
+  GdkDisplay *display = gdk_surface_get_display (GDK_SUBSURFACE (self)->parent);
+  gboolean done = FALSE;
+
+  /* We only return dmabuf information for GL textures.
+   * In other cases, we set n_planes to 0.
+   */
+  out_dmabuf->n_planes = 0;
+
+  if (GDK_IS_DMABUF_TEXTURE (texture))
+    {
+      const GdkDmabuf *dmabuf;
+
+      dmabuf = gdk_dmabuf_texture_get_dmabuf (GDK_DMABUF_TEXTURE (texture));
+      *out_fourcc = dmabuf->fourcc;
+      *out_premultiplied = gdk_memory_format_alpha (gdk_texture_get_format (texture)) == GDK_MEMORY_ALPHA_PREMULTIPLIED;
+      done = TRUE;
+    }
+  else if (GDK_IS_GL_TEXTURE (texture))
+    {
+      done = export_gl_texture_as_dmabuf (display, texture, out_dmabuf);
+      *out_fourcc = out_dmabuf->fourcc;
+      *out_premultiplied = gdk_memory_format_alpha (gdk_texture_get_format (texture)) == GDK_MEMORY_ALPHA_PREMULTIPLIED;
+    }
+
+  if (!done && GDK_DISPLAY_DEBUG_CHECK (display, FORCE_OFFLOAD))
+    {
+      GdkMemoryFormat format = gdk_texture_get_format (texture);
+
+      done = TRUE;
+      if (format == GDK_MEMORY_G8_B8R8_420)
+        *out_fourcc = DRM_FORMAT_NV12;
+      else if (gdk_memory_format_alpha (format) == GDK_MEMORY_ALPHA_OPAQUE)
+        *out_fourcc = DRM_FORMAT_RGBX8888;
+      else
+        *out_fourcc = DRM_FORMAT_RGBA8888;
+      *out_premultiplied = gdk_memory_format_alpha (format) == GDK_MEMORY_ALPHA_PREMULTIPLIED;
+    }
+
+  return done;
+}
+
 static struct wl_buffer *
-get_wl_buffer (GdkWaylandSubsurface *self,
-               GdkTexture           *texture)
+get_wl_buffer_from_info (GdkWaylandSubsurface *self,
+                         GdkTexture           *texture,
+                         GdkDmabuf            *dmabuf)
 {
   GdkDisplay *display = gdk_surface_get_display (GDK_SUBSURFACE (self)->parent);
   struct wl_buffer *buffer = NULL;
 
   if (GDK_IS_DMABUF_TEXTURE (texture))
     buffer = get_dmabuf_texture_wl_buffer (self, texture);
-  else if (GDK_IS_GL_TEXTURE (texture))
-    buffer = get_gl_texture_wl_buffer (self, texture);
+  else if (GDK_IS_GL_TEXTURE (texture) && dmabuf->n_planes > 0)
+    buffer = get_gl_texture_wl_buffer (self, texture, dmabuf);
 
-  if (GDK_DISPLAY_DEBUG_CHECK (display, FORCE_OFFLOAD))
-    {
-      if (!buffer)
-         buffer = get_shm_wl_buffer (self, texture);
-    }
+  if (!buffer && GDK_DISPLAY_DEBUG_CHECK (display, FORCE_OFFLOAD))
+    buffer = _gdk_wayland_shm_texture_get_wl_buffer (GDK_WAYLAND_DISPLAY (display), texture);
 
   return buffer;
 }
 
+/* }}} */
+/* {{{ Single-pixel buffer handling */
+
 static void
 sp_buffer_release (void             *data,
                    struct wl_buffer *buffer)
@@ -323,34 +354,8 @@ get_sp_buffer (GdkWaylandSubsurface *self)
   return buffer;
 }
 
-/* Note: The GdkDihedral transforms are *inverses* of the corresponding
- * wl_output_transform transforms.
- *
- * This is intentional: The GdkDihedral is the transform we want the
- * compositor to apply. set_buffer_transform is about *already transformed*
- * content. By telling the compositor that the content is already transformed
- * by the inverse of the GdkDihedral, we get it to apply the transform we want.
- */
-static inline enum wl_output_transform
-gdk_texture_transform_to_wl (GdkDihedral transform)
-{
-  return (enum wl_output_transform) transform;
-}
-
-static inline GdkDihedral
-wl_output_transform_to_gdk (enum wl_output_transform transform)
-{
-  return (GdkDihedral) transform;
-}
-
-G_STATIC_ASSERT ((int) WL_OUTPUT_TRANSFORM_NORMAL == (int) GDK_DIHEDRAL_NORMAL);
-G_STATIC_ASSERT ((int) WL_OUTPUT_TRANSFORM_90 == (int) GDK_DIHEDRAL_90);
-G_STATIC_ASSERT ((int) WL_OUTPUT_TRANSFORM_180 == (int) GDK_DIHEDRAL_180);
-G_STATIC_ASSERT ((int) WL_OUTPUT_TRANSFORM_270 == (int) GDK_DIHEDRAL_270);
-G_STATIC_ASSERT ((int) WL_OUTPUT_TRANSFORM_FLIPPED == (int) GDK_DIHEDRAL_FLIPPED);
-G_STATIC_ASSERT ((int) WL_OUTPUT_TRANSFORM_FLIPPED_90 == (int) GDK_DIHEDRAL_FLIPPED_90);
-G_STATIC_ASSERT ((int) WL_OUTPUT_TRANSFORM_FLIPPED_180 == (int) GDK_DIHEDRAL_FLIPPED_180);
-G_STATIC_ASSERT ((int) WL_OUTPUT_TRANSFORM_FLIPPED_270 == (int) GDK_DIHEDRAL_FLIPPED_270);
+/* }}} */
+/* {{{ Attach vfunc helpers */
 
 static void
 ensure_bg_surface (GdkWaylandSubsurface *self)
@@ -399,46 +404,9 @@ scaled_rect_is_integral (const graphene_rect_t *rect,
 }
 
 static gboolean
-gdk_wayland_subsurface_attach (GdkSubsurface         *sub,
-                               GdkTexture            *texture,
-                               const graphene_rect_t *source,
-                               const graphene_rect_t *dest,
-                               GdkDihedral            transform,
-                               const graphene_rect_t *background,
-                               gboolean               above,
-                               GdkSubsurface         *sibling)
+update_dest (GdkWaylandSubsurface  *self,
+             const graphene_rect_t *dest)
 {
-  GdkWaylandSubsurface *self = GDK_WAYLAND_SUBSURFACE (sub);
-  GdkWaylandSurface *parent = GDK_WAYLAND_SURFACE (sub->parent);
-  GdkWaylandDisplay *display = GDK_WAYLAND_DISPLAY (gdk_surface_get_display (sub->parent));
-  struct wl_buffer *buffer = NULL;
-  gboolean result = FALSE;
-  GdkWaylandSubsurface *sib = sibling ? GDK_WAYLAND_SUBSURFACE (sibling) : NULL;
-  gboolean will_be_above;
-  double scale;
-  graphene_rect_t device_rect;
-  gboolean has_background;
-  enum wl_output_transform tf;
-  gboolean dest_changed = FALSE;
-  gboolean source_changed = FALSE;
-  gboolean transform_changed = FALSE;
-  gboolean stacking_changed = FALSE;
-  gboolean needs_commit = FALSE;
-  gboolean background_changed = FALSE;
-  gboolean needs_bg_commit = FALSE;
-  gboolean color_state_changed = FALSE;
-
-  if (sibling)
-    will_be_above = sibling->above_parent;
-  else
-    will_be_above = above;
-
-  if (sub->parent == NULL)
-    {
-      g_warning ("Can't attach to destroyed subsurface %p", self);
-      return FALSE;
-    }
-
   if (self->dest.x != dest->origin.x ||
       self->dest.y != dest->origin.y ||
       self->dest.width != dest->size.width ||
@@ -448,38 +416,46 @@ gdk_wayland_subsurface_attach (GdkSubsurface         *sub,
       self->dest.y = dest->origin.y;
       self->dest.width = dest->size.width;
       self->dest.height = dest->size.height;
-      dest_changed = TRUE;
+      return TRUE;
     }
+  return FALSE;
+}
 
+static gboolean
+update_source (GdkWaylandSubsurface  *self,
+               const graphene_rect_t *source)
+{
   if (!gsk_rect_equal (&self->source, source))
     {
       self->source.origin.x = source->origin.x;
       self->source.origin.y = source->origin.y;
       self->source.size.width = source->size.width;
       self->source.size.height = source->size.height;
-      source_changed = TRUE;
+      return TRUE;
     }
+  return FALSE;
+}
+
+static gboolean
+update_transform (GdkWaylandSubsurface *self,
+                  GdkDihedral           transform)
+{
+  enum wl_output_transform tf;
 
   tf = gdk_texture_transform_to_wl (transform);
   if (self->transform != tf)
     {
       self->transform = tf;
-      transform_changed = TRUE;
+      return TRUE;
     }
+  return FALSE;
+}
 
-  if (sibling != gdk_subsurface_get_sibling (sub, above) ||
-      will_be_above != gdk_subsurface_is_above_parent (sub))
-    stacking_changed = TRUE;
-
-  if (self->texture == NULL)
-    {
-      dest_changed = TRUE;
-      source_changed = TRUE;
-      transform_changed = TRUE;
-      stacking_changed = TRUE;
-    }
-
-  scale = gdk_fractional_scale_to_double (&parent->scale);
+static gboolean
+update_background (GdkWaylandSubsurface  *self,
+                   const graphene_rect_t *background)
+{
+  gboolean background_changed;
 
   if (background)
     {
@@ -503,8 +479,91 @@ gdk_wayland_subsurface_attach (GdkSubsurface         *sub,
       self->bg_rect.height = 0;
     }
 
+  return background_changed;
+}
+
+/* }}} */
+/* {{{ The big, beautiful attach vfunc */
+
+static gboolean
+gdk_wayland_subsurface_attach (GdkSubsurface         *sub,
+                               GdkTexture            *texture,
+                               const graphene_rect_t *source,
+                               const graphene_rect_t *dest,
+                               GdkDihedral            transform,
+                               const graphene_rect_t *background,
+                               gboolean               above,
+                               GdkSubsurface         *sibling)
+{
+  GdkWaylandSubsurface *self = GDK_WAYLAND_SUBSURFACE (sub);
+  GdkWaylandSurface *parent = GDK_WAYLAND_SURFACE (sub->parent);
+  GdkWaylandDisplay *display = GDK_WAYLAND_DISPLAY (gdk_surface_get_display (sub->parent));
+  GdkDmabuf dmabuf = { .n_planes = 0, };
+  struct wl_buffer *buffer = NULL;
+  gboolean result = FALSE;
+  GdkWaylandSubsurface *sib = sibling ? GDK_WAYLAND_SUBSURFACE (sibling) : NULL;
+  gboolean will_be_above;
+  double scale;
+  graphene_rect_t device_rect;
+  gboolean has_background;
+  gboolean dest_changed = FALSE;
+  gboolean source_changed = FALSE;
+  gboolean transform_changed = FALSE;
+  gboolean stacking_changed = FALSE;
+  gboolean needs_commit = FALSE;
+  gboolean background_changed = FALSE;
+  gboolean needs_bg_commit = FALSE;
+  gboolean color_changed = FALSE;
+  gboolean transparent_changed = FALSE;
+  guint32 fourcc = 0;
+  gboolean premultiplied = TRUE;
+  gboolean was_transparent;
+  gboolean is_transparent;
+  GError *error = NULL;
+
+  if (sub->parent == NULL)
+    {
+      g_warning ("Can't attach to destroyed subsurface %p", self);
+      return FALSE;
+    }
+
+  if (sibling)
+    will_be_above = sibling->above_parent;
+  else
+    will_be_above = above;
+
+  if (sibling != gdk_subsurface_get_sibling (sub, above) ||
+      will_be_above != gdk_subsurface_is_above_parent (sub))
+    stacking_changed = TRUE;
+
+  dest_changed = update_dest (self, dest);
+  source_changed = update_source (self, source);
+  transform_changed = update_transform (self, transform);
+  if (self->texture)
+    {
+      was_transparent = gdk_memory_format_alpha (gdk_texture_get_format (self->texture)) != GDK_MEMORY_ALPHA_OPAQUE;
+    }
+  else
+    {
+      dest_changed = TRUE;
+      source_changed = TRUE;
+      transform_changed = TRUE;
+      stacking_changed = TRUE;
+      was_transparent = FALSE;
+    }
+
+  if (texture)
+    is_transparent = gdk_memory_format_alpha (gdk_texture_get_format (texture)) != GDK_MEMORY_ALPHA_OPAQUE;
+  else
+    is_transparent = TRUE;
+
+  transparent_changed = is_transparent != was_transparent;
+  background_changed = update_background (self, background);
+
   has_background = self->bg_rect.width > 0 && self->bg_rect.height > 0;
 
+  scale = gdk_fractional_scale_to_double (&parent->scale);
+
   if (!scaled_rect_is_integral (dest, 1, &device_rect))
     {
       GDK_DISPLAY_DEBUG (gdk_surface_get_display (sub->parent), OFFLOAD,
@@ -539,9 +598,7 @@ gdk_wayland_subsurface_attach (GdkSubsurface         *sub,
                          device_rect.size.width, device_rect.size.height,
                          scale);
     }
-  else if (!will_be_above &&
-           gdk_memory_format_alpha (gdk_texture_get_format (texture)) != GDK_MEMORY_ALPHA_OPAQUE &&
-           !has_background)
+  else if (!will_be_above && is_transparent && !has_background)
     {
       GDK_DISPLAY_DEBUG (gdk_surface_get_display (sub->parent), OFFLOAD,
                          "[%p] 🗙 Non-opaque texture (%dx%d) below",
@@ -555,70 +612,64 @@ gdk_wayland_subsurface_attach (GdkSubsurface         *sub,
                          "[%p] 🗙 Texture has background, but no single-pixel buffer support",
                          self);
     }
-  else if (self->color &&
-           !gdk_wayland_color_surface_can_set_color_state (self->color, gdk_texture_get_color_state (texture)))
+  else if (texture && texture != self->texture &&
+           !get_texture_info (self, texture, &fourcc, &premultiplied, &dmabuf))
     {
       GDK_DISPLAY_DEBUG (gdk_surface_get_display (sub->parent), OFFLOAD,
-                         "[%p] 🗙 Texture colorstate %s not supported",
+                         "[%p] 🗙 Texture type not supported or export failed",
+                         self);
+    }
+  else if (texture && texture != self->texture &&
+           !gdk_wayland_color_surface_can_set_color_state (self->color,
+                                                           gdk_texture_get_color_state (texture),
+                                                           fourcc, premultiplied,
+                                                           &error))
+    {
+      gdk_dmabuf_close_fds (&dmabuf);
+      GDK_DISPLAY_DEBUG (gdk_surface_get_display (sub->parent), OFFLOAD,
+                         "[%p] 🗙 Texture colorstate %s (%.4s, %s): %s",
                          self,
-                         gdk_color_state_get_name (gdk_texture_get_color_state (texture)));
+                         gdk_color_state_get_name (gdk_texture_get_color_state (texture)),
+                         (char *) &fourcc, premultiplied ? "premultiplied" : "straight",
+                         error->message);
+      g_error_free (error);
     }
   else
     {
-      gboolean was_transparent;
-
-      if (self->texture)
-        was_transparent = gdk_memory_format_alpha (gdk_texture_get_format (self->texture)) != GDK_MEMORY_ALPHA_OPAQUE;
-      else
-        was_transparent = FALSE;
-
       if (self->texture && texture)
-        color_state_changed = !gdk_color_state_equal (gdk_texture_get_color_state (self->texture),
-                                                      gdk_texture_get_color_state (texture));
+        {
+          color_changed = !gdk_color_state_equal (gdk_texture_get_color_state (self->texture),
+                                                  gdk_texture_get_color_state (texture)) ||
+                          self->fourcc != fourcc ||
+                          self->premultiplied != premultiplied;
+        }
       else
-        color_state_changed = TRUE;
+        {
+          color_changed = TRUE;
+        }
 
       if (g_set_object (&self->texture, texture))
         {
-          buffer = get_wl_buffer (self, texture);
-          if (buffer != NULL)
-            {
-              gboolean is_transparent;
-
-              is_transparent = gdk_memory_format_alpha (gdk_texture_get_format (texture)) != GDK_MEMORY_ALPHA_OPAQUE;
-              if (is_transparent != was_transparent)
-                {
-                  if (is_transparent)
-                    wl_surface_set_opaque_region (self->surface, NULL);
-                  else
-                    wl_surface_set_opaque_region (self->surface, self->opaque_region);
-                }
-
-              GDK_DISPLAY_DEBUG (gdk_surface_get_display (sub->parent), OFFLOAD,
-                                 "[%p] %s Attaching %s (%dx%d, %s) at %d %d %d %d%s%s%s",
-                                 self,
-                                 G_OBJECT_TYPE_NAME (texture),
-                                 will_be_above
-                                   ? (has_background ? "▲" : "△")
-                                   : (has_background ? "▼" : "▽"),
-                                 gdk_texture_get_width (texture),
-                                 gdk_texture_get_height (texture),
-                                 gdk_color_state_get_name (gdk_texture_get_color_state (texture)),
-                                 self->dest.x, self->dest.y,
-                                 self->dest.width, self->dest.height,
-                                 transform != GDK_DIHEDRAL_NORMAL ? " (" : "",
-                                 transform != GDK_DIHEDRAL_NORMAL ? gdk_dihedral_get_name (transform) : "",
-                                 transform != GDK_DIHEDRAL_NORMAL ? ")" : ""
-                                 );
-              result = TRUE;
-            }
-          else
-            {
-              GDK_DISPLAY_DEBUG (gdk_surface_get_display (sub->parent), OFFLOAD,
-                                 "[%p] 🗙 Failed to create wl_buffer for %s",
-                                 self,
-                                 G_OBJECT_TYPE_NAME (texture));
-            }
+          self->fourcc = fourcc;
+          self->premultiplied = premultiplied;
+          buffer = get_wl_buffer_from_info (self, texture, &dmabuf);
+          result = buffer != NULL;
+          if (result)
+            GDK_DISPLAY_DEBUG (gdk_surface_get_display (sub->parent), OFFLOAD,
+                               "[%p] %s Attaching %s (%dx%d, %s) at %d %d %d %d%s%s%s",
+                               self,
+                               G_OBJECT_TYPE_NAME (texture),
+                               will_be_above
+                                 ? (has_background ? "▲" : "△")
+                                 : (has_background ? "▼" : "▽"),
+                               gdk_texture_get_width (texture),
+                               gdk_texture_get_height (texture),
+                               gdk_color_state_get_name (gdk_texture_get_color_state (texture)),
+                               self->dest.x, self->dest.y,
+                               self->dest.width, self->dest.height,
+                               transform != GDK_DIHEDRAL_NORMAL ? " (" : "",
+                               transform != GDK_DIHEDRAL_NORMAL ? gdk_dihedral_get_name (transform) : "",
+                               transform != GDK_DIHEDRAL_NORMAL ? ")" : "");
         }
       else
         {
@@ -633,14 +684,19 @@ gdk_wayland_subsurface_attach (GdkSubsurface         *sub,
                                gdk_texture_get_height (texture),
                                self->dest.x, self->dest.y,
                                self->dest.width, self->dest.height);
-
-          buffer = NULL;
+          gdk_dmabuf_close_fds (&dmabuf);
           result = TRUE;
         }
     }
 
   if (result)
     {
+      if (transparent_changed)
+        {
+          wl_surface_set_opaque_region (self->surface, is_transparent ? NULL : self->opaque_region);
+          needs_commit = TRUE;
+        }
+
       if (transform_changed)
         {
           wl_surface_set_buffer_transform (self->surface, self->transform);
@@ -668,12 +724,18 @@ gdk_wayland_subsurface_attach (GdkSubsurface         *sub,
         {
           wl_surface_attach (self->surface, buffer, 0, 0);
 
-          if (self->color && color_state_changed)
-            gdk_wayland_color_surface_set_color_state (self->color, gdk_texture_get_color_state (texture));
+          if (color_changed)
+            {
+              GdkColorState *cs = gdk_texture_get_color_state (texture);
+              GDK_DISPLAY_DEBUG (gdk_surface_get_display (sub->parent), OFFLOAD,
+                                 "[%p] Setting color state %s",
+                                 self, gdk_color_state_get_name (cs));
+              gdk_wayland_color_surface_set_color_state (self->color, cs, fourcc, premultiplied);
+            }
 
           needs_commit = TRUE;
         }
-      
+
       if (buffer || transform_changed)
         {
           wl_surface_damage_buffer (self->surface,
@@ -716,11 +778,11 @@ gdk_wayland_subsurface_attach (GdkSubsurface         *sub,
               needs_bg_commit = TRUE;
             }
         }
-
-      result = TRUE;
     }
-  else
+  else /* !result */
     {
+      g_assert (buffer == NULL);
+
       if (g_set_object (&self->texture, NULL))
         {
           wl_surface_attach (self->surface, NULL, 0, 0);
@@ -737,29 +799,19 @@ gdk_wayland_subsurface_attach (GdkSubsurface         *sub,
 
   if (stacking_changed)
     {
-      if (sib)
-        {
-          if (above)
-            wl_subsurface_place_above (self->subsurface, sib->surface);
-          else
-            wl_subsurface_place_below (self->subsurface, sib->surface);
-        }
+      if (above)
+        wl_subsurface_place_above (self->subsurface,
+                                   sib ? sib->surface : GDK_WAYLAND_SURFACE (sub->parent)->display_server.wl_surface);
       else
-        {
-          if (above)
-            wl_subsurface_place_above (self->subsurface,
-                                       GDK_WAYLAND_SURFACE (sub->parent)->display_server.wl_surface);
-          else
-            wl_subsurface_place_below (self->subsurface,
-                                       GDK_WAYLAND_SURFACE (sub->parent)->display_server.wl_surface);
-        }
+        wl_subsurface_place_below (self->subsurface,
+                                   sib ? sib->surface : GDK_WAYLAND_SURFACE (sub->parent)->display_server.wl_surface);
       needs_commit = TRUE;
+    }
 
-      if (self->bg_attached)
-        {
-          wl_subsurface_place_below (self->bg_subsurface, self->surface);
-          needs_bg_commit = TRUE;
-        }
+  if (self->bg_attached)
+    {
+      wl_subsurface_place_below (self->bg_subsurface, self->surface);
+      needs_bg_commit = TRUE;
     }
 
   if (needs_commit)
@@ -774,6 +826,9 @@ gdk_wayland_subsurface_attach (GdkSubsurface         *sub,
   return result;
 }
 
+/* }}} */
+/* {{{ Other vfuncs */
+
 static void
 gdk_wayland_subsurface_detach (GdkSubsurface *sub)
 {
@@ -855,6 +910,36 @@ gdk_wayland_subsurface_get_background_rect (GdkSubsurface   *sub,
   return rect->size.width > 0 && rect->size.height > 0;
 }
 
+/* }}} */
+/* {{{ GObject boilerplate */
+
+G_DEFINE_TYPE (GdkWaylandSubsurface, gdk_wayland_subsurface, GDK_TYPE_SUBSURFACE)
+
+static void
+gdk_wayland_subsurface_init (GdkWaylandSubsurface *self)
+{
+}
+
+static void
+gdk_wayland_subsurface_finalize (GObject *object)
+{
+  GdkWaylandSubsurface *self = GDK_WAYLAND_SUBSURFACE (object);
+
+  g_clear_object (&self->texture);
+  g_clear_pointer (&self->frame_callback, wl_callback_destroy);
+  g_clear_pointer (&self->opaque_region, wl_region_destroy);
+  g_clear_pointer (&self->viewport, wp_viewport_destroy);
+  g_clear_pointer (&self->color, gdk_wayland_color_surface_free);
+  g_clear_pointer (&self->subsurface, wl_subsurface_destroy);
+  g_clear_pointer (&self->surface, wl_surface_destroy);
+  g_clear_pointer (&self->bg_viewport, wp_viewport_destroy);
+  g_clear_pointer (&self->bg_subsurface, wl_subsurface_destroy);
+  g_clear_pointer (&self->bg_surface, wl_surface_destroy);
+  g_clear_pointer (&self->idle_inhibitor, zwp_idle_inhibitor_v1_destroy);
+
+  G_OBJECT_CLASS (gdk_wayland_subsurface_parent_class)->finalize (object);
+}
+
 static void
 gdk_wayland_subsurface_class_init (GdkWaylandSubsurfaceClass *class)
 {
@@ -872,6 +957,9 @@ gdk_wayland_subsurface_class_init (GdkWaylandSubsurfaceClass *class)
   subsurface_class->get_background_rect = gdk_wayland_subsurface_get_background_rect;
 };
 
+/* }}} */
+/* {{{ API */
+
 static void
 frame_callback (void               *data,
                 struct wl_callback *callback,
@@ -929,8 +1017,7 @@ gdk_wayland_surface_create_subsurface (GdkSurface *surface)
   sub->subsurface = wl_subcompositor_get_subsurface (disp->subcompositor,
                                                      sub->surface,
                                                      impl->display_server.wl_surface);
-  if (disp->color)
-    sub->color = gdk_wayland_color_surface_new (disp->color, sub->surface, NULL, NULL);
+  sub->color = gdk_wayland_color_surface_new (disp->color, sub->surface, NULL, NULL);
 
   sub->viewport = wp_viewporter_get_viewport (disp->viewporter, sub->surface);
 
@@ -975,3 +1062,7 @@ gdk_wayland_subsurface_uninhibit_idle (GdkSubsurface *subsurface)
 
   g_clear_pointer (&sub->idle_inhibitor, zwp_idle_inhibitor_v1_destroy);
 }
+
+/* }}} */
+
+/* vim:set foldmethod=marker: */
